@@ -20,8 +20,7 @@ import (
 	"context"
 )
 
-type noopWriter struct {
-}
+type noopWriter struct{}
 
 func (p *noopWriter) Start(ctx context.Context) error {
 	return nil
@@ -38,6 +37,3 @@ func (p *noopWriter) TailMsgf(_ string, _ ...interface{}) {
 
 func (p *noopWriter) Stop() {
 }
-
-func (p *noopWriter) HasMore(bool) {
-}
